/*	$Id: qcpp.c,v 1.11 2011/10/24 06:29:02 nis Exp $	*/

/*-
 * Copyright (c) 2005 Shingo Nishioka.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#if ! defined lint
static char rcsid[] = "$Id: qcpp.c,v 1.11 2011/10/24 06:29:02 nis Exp $";
#endif

/* #include <config.h> */

#include <sys/types.h>

#include <getopt.h>
#include <libgen.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "qcpp.h"

#include <gifnoc.h>

static void unparse(char const *, struct cexpr *, FILE *);

static void c_unparse(struct cexpr *, FILE *);
static void tellline(int, FILE *);
static int elem_unparse(struct element *, FILE *, int);

static void funproto(struct cexpr *, FILE *);
static void elem_proto(struct element *, FILE *);

static void funbody(struct cexpr *, FILE *);
static int elem_args(struct element *, FILE *, int, int);
static void elem_body(struct element *, FILE *);
static void elem_cons(struct element *, FILE *, int);
static void cputs(char const *, int, int, FILE *);

char *progname = NULL;

static void
usage()
{
	fprintf(stderr, "usage: %s [-e] [-1] [-o ofile] [path]\n", progname);
	exit(1);
}

static int cc1 = 0;

int
main(argc, argv)
char *argv[];
{
	char const *path = NULL;
	struct cexpr *c;
	FILE *in, *out;
	char const *outf = NULL;
	int eflag = 0, ch;
	progname = basename(argv[0]);

	while ((ch = getopt(argc, argv, "e1o:")) != -1) {
		switch (ch) {
		case 'e':
			eflag = 1;
			break;
		case 'o':
			if (!(outf = strdup(optarg))) {
				perror("strdup");
				return 1;
			}
			break;
		case 'q':
			cc1 = 1;
			break;
		case '?':
			usage();
			break;
		}
	}

	argc -= optind;
	argv += optind;

	if (argc == 1) {
		path = argv[0];
		if (!(in = fopen(path, "r"))) {
			perror(path);
			return 1;
		}
	}
	else {
		in = stdin;
	}
	if (outf) {
		if (!(out = fopen(outf, "w"))) {
			perror(outf);
			return 1;
		}
	}
	else {
		out = stdout;
	}
	if (!(c = rqquote(in))) {
		fprintf(stderr, "%s: parse error\n", progname);
		return 1;
	}
	if (eflag == 0) {
		unparse(path, c, out);
	}
	return 0;
}

static void
unparse(path, c, stream)
char const *path;
struct cexpr *c;
FILE *stream;
{
	fputs("/* Generated by qcpp. DO NOT EDIT */\n", stream);
	fprintf(stream, "#line 1 \"%s\"\n", path);
	fputs("struct element;\n", stream);
	fputs("struct attribute;\n", stream);
	fputs("union content;\n", stream);
	funproto(c, stream);
	c_unparse(c, stream);
	funbody(c, stream);
}

static void
tellline(n, stream)
FILE *stream;
{
	if (cc1) {
		fprintf(stream, "\n# %d\n", n);
	}
	else {
		fprintf(stream, "\n#line %d\n", n);
	}
}

static void
c_unparse(c, stream)
struct cexpr *c;
FILE *stream;
{
	if (c && c->type == CCtoken) {
		tellline(c->line, stream);
	}
	for (; c; c = c->next) {
		switch (c->type) {
		case CCtoken:
			fputs(c->u.token, stream);
			break;
		case CQquote:
			fprintf(stream, QQCPFX "%p(", c->u.qquote);
			elem_unparse(c->u.qquote, stream, 0);
			fprintf(stream, ")");
			if (c->next && c->next->type == CCtoken) {
				tellline(c->next->line, stream);
			}
			break;
		}
	}
}

static int
elem_unparse(e, stream, comma)
struct element *e;
FILE *stream;
{
	struct attribute *a = e->attributes;
	union content **c = e->contents;
	struct cexpr *cx;

	if (a) {
		for (; a->type != 0; a++) {
			switch (a->type) {
			case AUnquote:
			case AUnquoteSplicing:
			case AUnquoteVal:
				if (comma) fputs(", ", stream);
				comma = 1;
				fputs("(", stream);
				cx = a->value.cexpr;
				c_unparse(cx->next, stream);
				fputs(")", stream);
				break;
			default:
				break;
			}
		}
	}

	if (c) {
		for (; *c; c++) {
			switch ((*c)->element.type) {
			case CElem:
				comma = elem_unparse(&(*c)->element, stream, comma);
				break;
			case CUnquote:
			case CUnquoteSplicing:
				if (comma) fputs(", ", stream);
				comma = 1;
				fputs("(", stream);
				cx = &(*c)->cexpr;
				c_unparse(cx->next, stream);
				fputs(")", stream);
				break;
			default:
				break;
			}
		}
	}
	return comma;
}

static void
funproto(c, stream)
struct cexpr *c;
FILE *stream;
{
	for (; c; c = c->next) {
		switch (c->type) {
		case CQquote:
			fprintf(stream, "static struct element *" QQCPFX "%p(", c->u.qquote);
			if (elem_args(c->u.qquote, stream, 0, 0) == 0) {
				fprintf(stream, "void");
			}
			fprintf(stream, ");\n");
			elem_proto(c->u.qquote, stream);
			break;
		default:
			break;
		}
	}
}

static void
elem_proto(e, stream)
struct element *e;
FILE *stream;
{
	struct attribute *a = e->attributes;
	union content **c = e->contents;
	struct cexpr *cx;

	if (a) {
		for (; a->type != 0; a++) {
			switch (a->type) {
			case AUnquote:
			case AUnquoteSplicing:
			case AUnquoteVal:
				cx = a->value.cexpr;
				funproto(cx->next, stream);
				break;
			default:
				break;
			}
		}
	}

	if (c) {
		for (; *c; c++) {
			switch ((*c)->element.type) {
			case CElem:
				elem_proto(&(*c)->element, stream);
				break;
			case CUnquote:
			case CUnquoteSplicing:
				cx = &(*c)->cexpr;
				funproto(cx->next, stream);
				break;
			default:
				break;
			}
		}
	}
}

static void
funbody(c, stream)
struct cexpr *c;
FILE *stream;
{
	for (; c; c = c->next) {
		switch (c->type) {
		case CQquote:
			fprintf(stream, "\n");
			fprintf(stream, "static struct element *\n" QQCPFX "%p(", c->u.qquote);
			if (elem_args(c->u.qquote, stream, 1, 0) == 0) {
				fprintf(stream, "void");
			}
			fprintf(stream, ")\n");
			fprintf(stream, "{\n");
			fprintf(stream, "\tstruct element *e;\n");
			fprintf(stream, "\tunion content *c, **sp;\n");
			fprintf(stream, "\tstruct attribute *ap, *aap;\n");
			fprintf(stream, "\tint ne;\n");
			fprintf(stream, "\tsp = NULL;\n");
			fprintf(stream, "\taap = NULL;\n");
			elem_args(c->u.qquote, stream, 2, 0);
			elem_cons(c->u.qquote, stream, 1);
			fprintf(stream, "\treturn e;\n");
			fprintf(stream, "}\n");
			elem_body(c->u.qquote, stream);
			break;
		default:
			break;
		}
	}
}

static int
elem_args(e, stream, name, comma)
struct element *e;
FILE *stream;
{
	struct attribute *a = e->attributes;
	union content **c = e->contents;

#define	ARG(type, cx) \
	if (name == 2) { \
		fprintf(stream, "\tif (!arg%p) return NULL;\n", (cx)); \
	} \
	else { \
		if (comma) fprintf(stream, ", "); \
		comma = 1; \
		fprintf(stream, #type); \
		if (name == 1) fprintf(stream, "arg%p", (cx)); \
	}

	if (a) {
		for (; a->type != 0; a++) {
			switch (a->type) {
			case AUnquote:
			case AUnquoteSplicing:
				ARG(struct attribute *, a->value.cexpr);
				break;
			case AUnquoteVal:
				ARG(char *, a->value.cexpr);
				break;
			default:
				break;
			}
		}
	}

	if (c) {
		for (; *c; c++) {
			switch ((*c)->element.type) {
			case CElem:
				comma = elem_args(&(*c)->element, stream, name, comma);
				break;
			case CUnquote:
				ARG(union content *, &(*c)->cexpr);
				break;
			case CUnquoteSplicing:
				ARG(union content **, &(*c)->cexpr);
				break;
			default:
				break;
			}
		}
	}
	return comma;
#undef ARG
}

static void
elem_body(e, stream)
struct element *e;
FILE *stream;
{
	struct attribute *a = e->attributes;
	union content **c = e->contents;
	struct cexpr *cx;

	if (a) {
		for (; a->type != 0; a++) {
			switch (a->type) {
			case AUnquote:
			case AUnquoteSplicing:
			case AUnquoteVal:
				cx = a->value.cexpr;
				funbody(cx->next, stream);
				break;
			default:
				break;
			}
		}
	}

	if (c) {
		for (; *c; c++) {
			switch ((*c)->element.type) {
			case CElem:
				elem_body(&(*c)->element, stream);
				break;
			case CUnquote:
			case CUnquoteSplicing:
				cx = &(*c)->cexpr;
				funbody(cx->next, stream);
				break;
			default:
				break;
			}
		}
	}
}

static void
elem_cons(e, stream, lev)
struct element *e;
FILE *stream;
{
#define	I	do { int l; for (l = 0; l < lev; l++) putc('\t', stream); } while (0)
	struct attribute *a = e->attributes;
	union content **c = e->contents;
	struct cexpr *cx;

	I; fprintf(stream, "if (!(c = alloc_content(CElem))) return NULL;\n");
	I; fprintf(stream, "e = &c->element;\n");
	I; fprintf(stream, "if (!(e->name = strdup(\"%s\"))) return NULL;\n", e->name);
	if (a) {
		int i, n, ne;
		for (n = 0; a[n].type != 0; n++) ;
		I; fprintf(stream, "ne = 0;\n");
		for (i = 0, ne = 0; i < n; i++) {
			struct attribute *ap = &a[i];
			switch (ap->type) {
			case AAttribute:
			case AUnquote:
			case AUnquoteVal:
				ne++;
				break;
			case AUnquoteSplicing:
				cx = ap->value.cexpr;
				I; fprintf(stream, "for (aap = arg%p; aap->name != NULL; aap++) ;\n", cx);
				I; fprintf(stream, "ne += aap - arg%p;\n", cx);
				break;
			default:
				break;
			}
		}
		I; fprintf(stream, "ne += %d;\n", ne + 1);
		I; fprintf(stream, "if (!(e->attributes = calloc(ne, sizeof *e->attributes))) return NULL;\n");
		I; fprintf(stream, "ap = e->attributes;\n");
		for (i = 0; i < n; i++) {
			struct attribute *ap = &a[i];
			switch (ap->type) {
			case AAttribute:
				I; fprintf(stream, "if (!(ap->name = strdup(\"");
				fputs(ap->name, stream);
				fprintf(stream, "\"))) return NULL;\n");
				I; fprintf(stream, "if (!(ap->value = strdup(");
				fputs(ap->value.string, stream);
				fprintf(stream, "))) return NULL;\n");
				I; fprintf(stream, "ap++;\n");
				break;
			case AUnquote:
				cx = ap->value.cexpr;
				I; fprintf(stream, "*ap++ = *arg%p;\n", cx);
				break;
			case AUnquoteSplicing:
				cx = ap->value.cexpr;
				I; fprintf(stream, "for (aap = arg%p; aap->name != NULL; *ap++ = *aap++) ;\n", cx);
				break;
			case AUnquoteVal:
				I; fprintf(stream, "if (!(ap->name = strdup(\"");
				fputs(ap->name, stream);
				fprintf(stream, "\"))) return NULL;\n");
				cx = ap->value.cexpr;
				I; fprintf(stream, "ap->value = (char *)arg%p;\n", cx);
				I; fprintf(stream, "ap++;\n");
				break;
			default:
				break;
			}
		}
		I; fprintf(stream, "ap->name = NULL;\n");
	}
	else {
		I; fprintf(stream, "e->attributes = NULL;\n");
	}
	if (c) {
		int i, n, ne;
		for (n = 0; c[n]; n++) ;
		I; fprintf(stream, "ne = 0;\n");
		for (i = 0, ne = 0; i < n; i++) {
			union content *cp = c[i];
			switch (cp->element.type) {
			case CElem:
			case CString:
			case CComment:
			case CUnquote:
				ne++;
				break;
			case CUnquoteSplicing:
				cx = &cp->cexpr;
				I; fprintf(stream, "for (sp = arg%p; *sp; sp++) ;\n", cx);
				I; fprintf(stream, "ne += sp - arg%p;\n", cx);
				break;
			default:
				break;
			}
		}
		I; fprintf(stream, "ne += %d;\n", ne + 1);
		I; fprintf(stream, "if (!(e->contents = calloc(ne, sizeof *e->contents))) return NULL;\n");
		I; fprintf(stream, "{\n");
		lev++;
		I; fprintf(stream, "struct element *p = e;\n");
		I; fprintf(stream, "union content **cp = p->contents;\n");
		for (i = 0; i < n; i++) {
			union content *cp = c[i];
			switch (cp->element.type) {
			case CElem:
				elem_cons(&cp->element, stream, lev);
				I; fprintf(stream, "*cp++ = (union content *)e;\n");
				break;
			case CString:
				I; fprintf(stream, "if (!(*cp = alloc_content(CString))) return NULL;\n");
				I; fprintf(stream, "(*cp)->chardata.value = ");
				fputs(cp->chardata.value, stream);
				fprintf(stream, ";\n");
				I; fprintf(stream, "cp++;\n");
				break;
			case CComment:
				I; fprintf(stream, "if (!(*cp = alloc_content(CComment))) return NULL;\n");
				I; fprintf(stream, "(*cp)->misc.string = strdup(\")");
				cputs(cp->misc.string, 2, strlen(cp->misc.string) - 2, stream);
				fprintf(stream, "\");\n");
				I; fprintf(stream, "cp++;\n");
				break;
			case CUnquote:
				cx = &cp->cexpr;
				I; fprintf(stream, "*cp++ = arg%p;\n", cx);
				break;
			case CUnquoteSplicing:
				cx = &cp->cexpr;
				I; fprintf(stream, "for (sp = arg%p; *sp; *cp++ = *sp++) ;\n", cx);
			default:
				break;
			}
		}
		I; fprintf(stream, "*cp = NULL;\n");
		I; fprintf(stream, "e = p;\n");
		lev--;
		I; fprintf(stream, "}\n");
	}
	else {
		I; fprintf(stream, "e->contents = NULL;\n");
	}
}

static void
cputs(s, start, end, stream)
char const *s;
FILE *stream;
{
	int i;
	for (i = start; i < end; i++) {
		int c = s[i];
		if (c == '"' || c == '\\') {
			putc('\\', stream);
		}
		putc(c, stream);
	}
}
