/*	$Id: qqlex.c,v 1.8 2009/07/30 00:10:05 nis Exp $	*/

/*-
 * Copyright (c) 2005 Shingo Nishioka.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#if ! defined lint
static char rcsid[] = "$Id: qqlex.c,v 1.8 2009/07/30 00:10:05 nis Exp $";
#endif

/* #include <config.h> */

#include <sys/types.h>

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "qcpp.h"
#include "y.tab.h"

#include <gifnoc.h>

struct {
	FILE *stream;
	int line, column, tokline, lastchar, pushback;
	char *path;
	struct {
		int line;
		char *path;
	} next;
} str;

static int mygetc(void);
static void myungetc(int);

void
init_mygetc(stream)
FILE *stream;
{
	str.stream = stream;
	str.line = 1;
	str.column = 1;
	str.path = NULL;
	str.lastchar = EOF;
	str.pushback = EOF;
	str.next.line = -1;
}

static int
mygetc()
{
	int c;
	if (str.pushback != EOF) {
		c = str.pushback;
		str.pushback = EOF;
		return c;
	}
	c = getc(str.stream);
	str.column++;
	if (c == '\n') {
		if (str.next.line != -1) {
			str.line = str.line;
			str.next.line = -1;
		}
		else {
			str.line++;
		}
		str.column = 1;
	}
	str.lastchar = c;
	return c;
}

static void
myungetc(c)
{
	if (c == EOF) return;
	if (str.pushback != EOF) {
		fprintf(stderr, "internal error: too many pushback char\n");
		exit(1);
	}
	str.pushback = c;
}

int
yytokline()
{
	return str.tokline;
}

int
yyline()
{
	return str.line;
}

int
yycolumn()
{
	return str.column;
}

int
yylastchar()
{
	return str.lastchar;
}

void
yysetline(n)
{
	str.next.line = n;
}

int
yylex()
{
	static char *buf = NULL;
	static size_t bufsize = 0;
	int c, i;

#define XB(s) \
	if ((s) >= bufsize) { \
		void *new; \
		size_t newsize; \
		newsize = bufsize + 256; \
		if (!(new = realloc(buf, newsize))) { \
			perror("realloc"); \
			goto error; \
		} \
		buf = new; \
		bufsize = newsize; \
	}

	str.tokline = str.line;
	c = mygetc();

	if (c == '"') {
		i = 0;
		XB(i);
		buf[i++] = c;
		for (; (c = mygetc()) != EOF && c != '"'; ) {
			XB(i);
			buf[i++] = c;
			if (c == '\\') {
				c = mygetc();
				XB(i);
				buf[i++] = c;
			}
		}
		if (c == EOF) {
			fprintf(stderr, "EOF in string");
			goto error;
		}
		XB(i + 1);
		buf[i++] = c;
		buf[i] = '\0';
		if (!(yylval.name = strdup(buf))) {
			perror("strdup");
			goto error;
		}
		return STRING;
	}
	else if (isalpha(c)) {
		i = 0;
		XB(i);
		buf[i++] = c;
		for (; (c = mygetc()) != EOF && (isalnum(c) || c == '_' || c == '-' || c == ':'); ) {
			XB(i);
			buf[i++] = c;
		}
		XB(i);
		buf[i] = '\0';
		myungetc(c);
		if (!(yylval.name = strdup(buf))) {
			perror("strdup");
			goto error;
		}
		return ID;
	}
	else if (isdigit(c)) {
		i = 0;
		XB(i);
		buf[i++] = c;
		for (; (c = mygetc()) != EOF && isdigit(c); ) {
			XB(i);
			buf[i++] = c;
		}
		XB(i);
		buf[i] = '\0';
		myungetc(c);
		if (!(yylval.name = strdup(buf))) {
			perror("strdup");
			goto error;
		}
		return INUM;
	}
	if (c == '\'') {
		i = 0;
		XB(i);
		buf[i++] = c;
		for (; (c = mygetc()) != EOF && c != '\''; ) {
			XB(i);
			buf[i++] = c;
			if (c == '\\') {
				c = mygetc();
				XB(i);
				buf[i++] = c;
			}
		}
		if (c == EOF) {
			fprintf(stderr, "EOF in character constant");
			goto error;
		}
		XB(i + 1);
		buf[i++] = c;
		buf[i] = '\0';
		if (!(yylval.name = strdup(buf))) {
			perror("strdup");
			goto error;
		}
		return CTOKEN;
	}
	else if (c == '/') {
		c = mygetc();
		if (c == '*') {
			int pc;

			XB(2);
			buf[0] = '/';
			buf[1] = '*';
			pc = buf[2] = mygetc();
			if (pc == EOF) {
				fprintf(stderr, "EOF in comment");
				goto error;
			}
			for (i = 3; (c = mygetc()) != EOF && !(pc == '*' && c == '/'); ) {
				XB(i);
				buf[i++] = c;
				pc = c;
			}
			if (c == EOF) {
				fprintf(stderr, "EOF in comment");
				goto error;
			}
			XB(i + 1);
			buf[i++] = c;
			buf[i] = '\0';
			if (!(yylval.name = strdup(buf))) {
				perror("strdup");
				goto error;
			}
			return COMMENT;
		}
		else if (c == '/') {
			XB(1);
			buf[0] = '/';
			buf[1] = '/';
			for (i = 2; (c = mygetc()) != EOF && c != '\n'; ) {
				XB(i);
				buf[i++] = c;
			}
			if (c == EOF) {
				fprintf(stderr, "EOF in comment");
				goto error;
			}
			XB(i + 1);
			buf[i++] = c;
			buf[i] = '\0';
			if (!(yylval.name = strdup(buf))) {
				perror("strdup");
				goto error;
			}
			return COMMENT;
		}
		else {
			XB(1);
			buf[0] = '/';
			buf[1] = '\0';
			myungetc(c);
			if (!(yylval.name = strdup(buf))) {
				perror("strdup");
				goto error;
			}
			return CTOKEN;
		}
	}
	else if (c == '(' || c == ')' || c == ',' || c == '=' || /* valid anywhere */
		 c == '{' || c == '}' || c == ';' || /* valid only toplevel */
		 c == '#' ||
		 c == '@' || c == '`') {	/* permanently illegal char */
		return c;
	}
	else if (c == ' ' || c == '\t' || c == '\n' || c == '\r') {
		XB(1);
		buf[0] = c;
		buf[1] = '\0';
		if (!(yylval.name = strdup(buf))) {
			perror("strdup");
			goto error;
		}
		return SPACE;
	}
	else if (c == EOF) {
		return 0;
	}

	XB(1);
	buf[0] = c;
	buf[1] = '\0';
	if (!(yylval.name = strdup(buf))) {
		perror("strdup");
		goto error;
	}
	return CTOKEN;

error:
	exit(1);
}
